from desdeo_tools.interaction.request import (
    PrintRequest,
    SimplePlotRequest,
    ReferencePointPreference,
)

from desdeo_tools.interaction.validators import(
    validate_ref_point_data_type,
    validate_ref_point_dimensions,
    validate_ref_point_with_ideal,
    validate_ref_point_with_ideal_and_nadir,
    validate_with_ref_point_nadir
)